@@ -16,10 +16,10 @@
 #ifndef _STACK_H
 #define _STACK_H
 
-struct _call_stack_t;
-typedef struct _call_stack_t call_stack_t;
-struct _call_frame_t;
-typedef struct _call_frame_t call_frame_t;
+struct _call_stack;
+typedef struct _call_stack call_stack_t;
+struct _call_frame;
+typedef struct _call_frame call_frame_t;
 struct call_pool;
 typedef struct call_pool call_pool_t;
 
@@ -29,10 +29,10 @@ typedef struct call_pool call_pool_t;
 #include "dict.h"
 #include "list.h"
 #include "common-utils.h"
-#include "globals.h"
 #include "lkowner.h"
 #include "client_t.h"
 #include "libglusterfs-messages.h"
+#include "timespec.h"
 
 #define NFS_PID 1
 #define LOW_PRIO_PROC_PID -1
@@ -56,12 +56,13 @@ struct call_pool {
                 } all_stacks;
         };
         int64_t                     cnt;
+        gf_atomic_t                 total_count;
         gf_lock_t                   lock;
         struct mem_pool             *frame_mem_pool;
         struct mem_pool             *stack_mem_pool;
 };
 
-struct _call_frame_t {
+struct _call_frame {
         call_stack_t *root;        /* stack root */
         call_frame_t *parent;      /* previous BP */
         struct list_head frames;
@@ -73,9 +74,9 @@ struct _call_frame_t {
         void         *cookie;      /* unique cookie */
         gf_boolean_t  complete;
 
-        glusterfs_fop_t op;
-        struct timeval begin;      /* when this frame was created */
-        struct timeval end;        /* when this frame completed */
+        glusterfs_fop_t  op;
+        struct timespec  begin;      /* when this frame was created */
+        struct timespec  end;        /* when this frame completed */
         const char      *wind_from;
         const char      *wind_to;
         const char      *unwind_from;
@@ -84,7 +85,7 @@ struct _call_frame_t {
 
 #define SMALL_GROUP_COUNT 128
 
-struct _call_stack_t {
+struct _call_stack {
         union {
                 struct list_head      all_frames;
                 struct {
@@ -113,9 +114,14 @@ struct _call_stack_t {
 
         int32_t                       op;
         int8_t                        type;
-        struct timeval                tv;
+        struct timespec               tv;
         xlator_t                     *err_xl;
         int32_t                       error;
+
+        uint32_t flags; /* use it wisely, think of it as a mechanism to
+                           send information over the wire too */
+        struct timespec ctime; /* timestamp, most probably set at
+                                  creation of stack. */
 };
 
 
@@ -130,18 +136,17 @@ struct _call_stack_t {
 
 
 struct xlator_fops;
+void gf_update_latency (call_frame_t *frame);
 
-void
-gf_latency_begin (call_frame_t *frame, void *fn);
-
-void
-gf_latency_end (call_frame_t *frame);
 
 static inline void
 FRAME_DESTROY (call_frame_t *frame)
 {
         void *local = NULL;
 
+        if (frame->root->ctx->measure_latency)
+                gf_update_latency (frame);
+
         list_del_init (&frame->frames);
         if (frame->local) {
                 local = frame->local;
@@ -226,6 +231,18 @@ STACK_RESET (call_stack_t *stack)
         } while (0);                                                   \
 
 
+/* NOTE: make sure to keep this as an macro, mainly because, we need 'fn'
+   field here to be the proper fn ptr, so its address is valid entry in
+   'xlator_fops' struct.
+   To understand this, check the `xlator.h:struct xlator_fops`, and then
+   see a STACK_WIND call, which generally calls `subvol->fops->fop`, so
+   the address offset should give the index */
+
+/* +1 is required as 0 means NULL fop, and we don't have a variable for it */
+#define get_fop_index_from_fn(xl, fn)                                   \
+        (1 + (((long)&(fn) - (long)&((xl)->fops->stat)) / sizeof (void *)))
+
+
 /* make a call without switching frames */
 #define STACK_WIND_TAIL(frame, obj, fn, params ...)                     \
         do {                                                            \
@@ -242,6 +259,14 @@ STACK_RESET (call_stack_t *stack)
                               "winding from %s to %s",                  \
                               frame->root, old_THIS->name,              \
                               THIS->name);                              \
+                /* Need to capture counts at leaf node */               \
+                if (!next_xl->children) {                               \
+                        int op = get_fop_index_from_fn((next_xl), (fn)); \
+                        GF_ATOMIC_INC (next_xl->stats.total.metrics[op].fop); \
+                        GF_ATOMIC_INC (next_xl->stats.interval.metrics[op].fop); \
+                        GF_ATOMIC_INC (next_xl->stats.total.count);     \
+                        GF_ATOMIC_INC (next_xl->stats.interval.count);  \
+                }                                                       \
                 next_xl_fn (frame, next_xl, params);                    \
                 THIS = old_THIS;                                        \
         } while (0)
@@ -296,7 +321,12 @@ STACK_RESET (call_stack_t *stack)
                               frame->root, old_THIS->name,              \
                               THIS->name);                              \
                 if (obj->ctx->measure_latency)                          \
-                        gf_latency_begin (_new, fn);                    \
+                        timespec_now (&_new->begin);                    \
+                _new->op = get_fop_index_from_fn ((_new->this), (fn));  \
+                GF_ATOMIC_INC (obj->stats.total.metrics[_new->op].fop); \
+                GF_ATOMIC_INC (obj->stats.interval.metrics[_new->op].fop); \
+                GF_ATOMIC_INC (obj->stats.total.count);                 \
+                GF_ATOMIC_INC (obj->stats.interval.count);              \
                 fn (_new, obj, params);                                 \
                 THIS = old_THIS;                                        \
         } while (0)
@@ -305,9 +335,9 @@ STACK_RESET (call_stack_t *stack)
 #define STACK_UNWIND STACK_UNWIND_STRICT
 
 /* return from function in type-safe way */
-#define STACK_UNWIND_STRICT(op, frame, op_ret, op_errno, params ...)    \
+#define STACK_UNWIND_STRICT(fop, frame, op_ret, op_errno, params ...)   \
         do {                                                            \
-                fop_##op##_cbk_t      fn = NULL;                        \
+                fop_##fop##_cbk_t      fn = NULL;                       \
                 call_frame_t *_parent = NULL;                           \
                 xlator_t     *old_THIS = NULL;                          \
                                                                         \
@@ -316,30 +346,30 @@ STACK_RESET (call_stack_t *stack)
                                 LG_MSG_FRAME_ERROR, "!frame");          \
                         break;                                          \
                 }                                                       \
-                if (op_ret < 0) {                                       \
+                if ((op_ret) < 0) {                                     \
                         gf_msg_debug ("stack-trace", op_errno,          \
                                       "stack-address: %p, "             \
                                       "%s returned %d error: %s",       \
                                       frame->root, THIS->name,          \
-                                      (int32_t)op_ret,                  \
+                                      (int32_t)(op_ret),                \
                                       strerror(op_errno));              \
                 } else {                                                \
                         gf_msg_trace ("stack-trace", 0,                 \
                                       "stack-address: %p, "             \
                                       "%s returned %d",                 \
                                       frame->root, THIS->name,          \
-                                      (int32_t)op_ret);                 \
+                                      (int32_t)(op_ret));               \
                 }                                                       \
-                fn = (fop_##op##_cbk_t )frame->ret;                     \
+                fn = (fop_##fop##_cbk_t)frame->ret;                     \
                 _parent = frame->parent;                                \
                 LOCK(&frame->root->stack_lock);                         \
                 {                                                       \
                         _parent->ref_count--;                           \
-                        if (op_ret < 0 &&                               \
-                            op_errno != frame->root->error) {           \
+                        if ((op_ret) < 0 &&                             \
+                            (op_errno) != frame->root->error) {         \
                                 frame->root->err_xl = frame->this;      \
-                                frame->root->error = op_errno;          \
-                        } else if (op_ret == 0) {                       \
+                                frame->root->error = (op_errno);        \
+                        } else if ((op_ret) == 0) {                     \
                                 frame->root->err_xl = NULL;             \
                                 frame->root->error = 0;                 \
                         }                                               \
@@ -349,8 +379,16 @@ STACK_RESET (call_stack_t *stack)
                 THIS = _parent->this;                                   \
                 frame->complete = _gf_true;                             \
                 frame->unwind_from = __FUNCTION__;                      \
-                if (frame->this->ctx->measure_latency)                  \
-                        gf_latency_end (frame);                         \
+                if (frame->this->ctx->measure_latency) {                \
+                        timespec_now (&frame->end);                     \
+                        /* required for top most xlator */              \
+                        if (_parent->ret == NULL)                       \
+                                timespec_now (&_parent->end);           \
+                }                                                       \
+                if (op_ret < 0) {                                       \
+                        GF_ATOMIC_INC (THIS->stats.total.metrics[frame->op].cbk); \
+                        GF_ATOMIC_INC (THIS->stats.interval.metrics[frame->op].cbk); \
+                }                                                       \
                 fn (_parent, frame->cookie, _parent->this, op_ret,      \
                     op_errno, params);                                  \
                 THIS = old_THIS;                                        \
@@ -363,7 +401,7 @@ call_stack_alloc_groups (call_stack_t *stack, int ngrps)
 	if (ngrps <= SMALL_GROUP_COUNT) {
 		stack->groups = stack->groups_small;
 	} else {
-		stack->groups_large = GF_CALLOC (sizeof (gid_t), ngrps,
+		stack->groups_large = GF_CALLOC (ngrps, sizeof (gid_t),
 						 gf_common_mt_groups_t);
 		if (!stack->groups_large)
 			return -1;
@@ -424,13 +462,24 @@ copy_frame (call_frame_t *frame)
         newstack->uid = oldstack->uid;
         newstack->gid = oldstack->gid;
         newstack->pid = oldstack->pid;
-        newstack->ngrps = oldstack->ngrps;
         newstack->op  = oldstack->op;
         newstack->type = oldstack->type;
 	if (call_stack_alloc_groups (newstack, oldstack->ngrps) != 0) {
 		mem_put (newstack);
 		return NULL;
 	}
+        if (!oldstack->groups) {
+                gf_msg_debug ("stack", EINVAL, "groups is null (ngrps: %d)",
+                              oldstack->ngrps);
+                /* Considering 'groups' is NULL, set ngrps to 0 */
+                oldstack->ngrps = 0;
+
+                if (oldstack->groups_large)
+                        oldstack->groups = oldstack->groups_large;
+                else
+                        oldstack->groups = oldstack->groups_small;
+        }
+        newstack->ngrps = oldstack->ngrps;
         memcpy (newstack->groups, oldstack->groups,
                 sizeof (gid_t) * oldstack->ngrps);
         newstack->unique = oldstack->unique;
@@ -439,10 +488,7 @@ copy_frame (call_frame_t *frame)
         newstack->ctx = oldstack->ctx;
 
         if (newstack->ctx->measure_latency) {
-                if (gettimeofday (&newstack->tv, NULL) == -1)
-                        gf_msg ("stack", GF_LOG_ERROR, errno,
-                                LG_MSG_GETTIMEOFDAY_FAILED,
-                                "gettimeofday () failed.");
+                timespec_now (&newstack->tv);
                 memcpy (&newframe->begin, &newstack->tv,
                         sizeof (newstack->tv));
         }
@@ -456,10 +502,12 @@ copy_frame (call_frame_t *frame)
                 newstack->pool->cnt++;
         }
         UNLOCK (&oldstack->pool->lock);
+        GF_ATOMIC_INC (newstack->pool->total_count);
 
         return newframe;
 }
 
+void call_stack_set_groups (call_stack_t *stack, int ngrps, gid_t **groupbuf_p);
 void gf_proc_dump_pending_frames(call_pool_t *call_pool);
 void gf_proc_dump_pending_frames_to_dict (call_pool_t *call_pool,
                                           dict_t *dict);
